from pathlib import Path
from functools import partial
import logging
# logging.basicConfig(level=logging.INFO)
# logging.basicConfig(level=logging.DEBUG)

from rich import print
from rich.pretty import pprint
pprint = partial(pprint, max_length=500)

from chocolatey import Chocolatey
choco = Chocolatey()

sources = choco.sources()
print("SOURCES:") ; print(sources)

pkg_info = choco.info(pkg_id="chocolatey")
print("PACKAGE INFO:") ; print(pkg_info)

input("Press a key...")



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

__all__ = ('issubtype', 'issequence', 'isiterable', 'remove_all', 'pushd')

from typing import Any, Sequence, Iterable, List
from collections import abc
import contextlib
String = str


def issubtype(x: Any, t: Any) -> bool:
    return isinstance(x, type) and issubclass(x, t)


def issequence(x: Any) -> bool:
    return (isinstance(x, (Sequence, abc.Sequence))
            and not isinstance(x, (bytes, str, String)))


def isiterable(x: Any) -> bool:
    return (isinstance(x, (Iterable, abc.Iterable))
            and not isinstance(x, (bytes, str, String)))


def remove_all(list: List, value: Any) -> None:  # noqa: A002
    list[:] = (item for item in list if item != value)


@contextlib.contextmanager
def pushd(path):
    import os
    curr_dir = os.getcwd()
    os.chdir(str(path) if isinstance(path, os.PathLike) else path)
    try:
        yield
    finally:
        os.chdir(curr_dir)


del contextlib



# Copyright (c) 2022 Adam Karpierz
# SPDX-License-Identifier: Zlib

from __future__ import annotations

"""Chocolatey API"""

from typing import Any, Union, List, Dict
from dataclasses import dataclass
from collections import defaultdict
from pathlib import Path
import tempfile
import shutil
import textwrap
# import enum
# from rich import print

from public import public
from nocasedict import NocaseDict
import regex as re

from ._run import run
from ._chocolatey_cmd import ChocolateyCmd


@public
@dataclass
class version_info:
    major:  int = 0
    minor:  int = 0
    micro:  int = 0
    serial: int = 0


@public
@dataclass
class Package:
    id: str  # noqa: A003
    version: str


@public
@dataclass
class PackageOutdated(Package):
    available_version: str = None
    pinned: bool = False

    def __post_init__(self):
        """Post-init"""
        self.pinned = _str2bool("pinned", self.pinned)


@public
@dataclass
class PackageInfo(Package):
    description: str = ""
    title: str = ""
    summary: str = ""
    # additional
    published: str = ""


@public
@dataclass
class Config:
    name: str
    value: Union[str, bool] | None = None
    description: str = ""

    def __post_init__(self):
        """Post-init"""
        self.value = _str2none("value", self.value)
        self.value = _str2bool("value", self.value, with_check=False)


@public
@dataclass
class Source:
    name: str
    value: str = ""
    disabled: bool = False
    user: str | None = None
    password: str | None = None
    priority: int = 0
    bypass_proxy: bool = False
    self_service: bool = False
    admin_only: bool = False

    def __post_init__(self):
        """Post-init"""
        self.disabled     = _str2bool("disabled", self.disabled)
        self.user         = _str2none("user", self.user)
        self.password     = _str2none("password", self.password)
        self.priority     = _str2int("priority", self.priority)
        self.bypass_proxy = _str2bool("bypass_proxy", self.bypass_proxy)
        self.self_service = _str2bool("self_service", self.self_service)
        self.admin_only   = _str2bool("admin_only", self.admin_only)


@public
@dataclass
class Feature:
    name: str
    enabled: bool = False
    description: str = ""
    # set_explicitly: bool = False

    def __post_init__(self):
        """Post-init"""
        self.enabled = _str2bool("enabled", self.enabled,
                                 literals=("enabled", "disabled"))


@public
@dataclass
class ApiKey:
    source: str
    info: str


@public
@dataclass
class Template:
    name: str
    version: str


@public
class Chocolatey:
    """Chocolatey API"""

    _allow_multiple = False

    def __new__(cls, source: str = None):
        self = super().__new__(cls)
        self.source = source
        self.cmd    = ChocolateyCmd(self.source)
        return self

    ### High-level API ###

    @classmethod
    def setup(cls):
        """Setup chocolatey software."""
        setup_dir = Path(__file__).resolve().parent/"choco-setup"
        choco_pkg = next(setup_dir.glob("chocolatey.*.nupkg"))
        # Install chocolatey
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file = Path(temp_dir)/"chocolatey.zip"
            shutil.copy2(str(choco_pkg), str(zip_file))
            run(shutil.which("powershell.exe"), "-ExecutionPolicy", "Bypass",
                setup_dir/"install.ps1", "-ChocolateyDownloadUrl", zip_file)

    @property
    def version(self) -> str:
        """Gets the Chocolatey version."""
        try:
            output = self.cmd.choco(version=True, limit_output=True,
                                    source=False, **self._capture_output)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return output.stdout.strip()

    @property
    def version_info(self) -> version_info:
        """Gets the Chocolatey version info."""
        parts = [int(item) for item in self.version.split(".")]
        parts[len(parts):] =  (4 - len(parts)) * [0]
        return version_info(*parts)

    def help(self, *, command: str = None) -> str:  # noqa: A003
        """Gets the help information for choco and choco commands."""
        try:
            if not command:
                output = self.cmd.help(limit_output=True,
                                       **self._capture_output)
            else:
                output = self.cmd.choco(command, help=True,
                                        limit_output=True,
                                        **self._capture_output)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return output.stdout.lstrip().replace("\r\n", "\n")

    # run_silent = partial(subprocess.run, stdout=open(os.devnull, 'wb'))
    # FIXME: look at python_vagrant to achieve hide of out and/or err stream

    def installed(self, *filters, all_versions=True,
                  **kwargs) -> Dict[str, List[Package]]:
        """Retrieves a list of locally installed packages."""
        self._omit_args(kwargs, "limit_output", "local_only",
                        "verbose", "detail", "detailed", "idonly", "id_only")
        try:
            output = self.cmd.list(*filters, limit_output=True, local_only=True,
                                   all_versions=all_versions, source=False,
                                   **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._packages(output.stdout)

    def outdated(self, *, ignore_pinned=True, ignore_unfound=True,
                 **kwargs) -> Dict[str, List[PackageOutdated]]:
        """Retrieves information about packages that are outdated."""
        self._omit_args(kwargs, "limit_output",
                        "verbose", "detail", "detailed", "idonly", "id_only")
        try:
            output = self.cmd.outdated(limit_output=True,
                                       ignore_pinned=ignore_pinned,
                                       ignore_unfound=ignore_unfound,
                                       **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        packages = self._packages(output.stdout, klass=PackageOutdated)
        for pkg_id, val in list(packages.items()):
            pkgs = [val] if isinstance(val, PackageOutdated) else val
            for idx, pkg in enumerate(pkgs[:]):
                if pkg.version == pkg.available_version: del pkgs[idx]
            if not pkgs: del packages[pkg_id]
        return packages

    def search(self, filter=False, *, all_versions=False,  # noqa: A002
               exact=False, **kwargs) -> Dict[str, List[Package]]:
        """Searches remote packages."""
        self._omit_args(kwargs, "limit_output", "page", "page_size",
                        "verbose", "detail", "detailed", "idonly", "id_only")
        out  = "" ; page = 0
        while True:
            try:
                arg = [filter] if filter is not False else []
                output = self.cmd.search(*arg, limit_output=True, page=page,
                                         all_versions=all_versions, exact=exact,
                                         **self._capture_output, **kwargs)
            except run.CalledProcessError as exc:
                self._handle_exception(exc)
            if not output.stdout: break
            out += output.stdout ; page += 1
            if exact: break
        return self._packages(out, allow_multiple=all_versions)

    def info(self, *, pkg_id: str, local_only=False, **kwargs) \
            -> PackageInfo | None:
        """Retrieves package information."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.info(pkg_id, limit_output=True,
                                   local_only=local_only,
                                   **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        # if not found, returns None
        if not output.stdout.strip():
            return None  # pragma: no cover
        packages = self._packages(output.stdout, klass=PackageInfo)
        if not packages:
            return None  # pragma: no cover
        pkg_info = list(packages.values())[0]
        try:
            output = self.cmd.info(pkg_id,
                                   local_only=local_only,
                                   **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._info(output.stdout, pkg_info)

    def export(self, output_file_path=False, *, include_version=True,
               **kwargs) -> None:
        """Exports list of currently installed packages."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.export(output_file_path=output_file_path,
                            include_version=include_version,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def install(self, *pkg_ids, yes=True, **kwargs) -> None:
        """Installs packages using configured sources."""
        if not pkg_ids:
            raise Chocolatey.TypeError("install() "
                                       "missing at least 1 required positional argument")
        self._omit_args(kwargs, "yes")  # , "verbose")
        try:
            self.cmd.install(*pkg_ids, yes=yes, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def upgrade(self, *pkg_ids, install_if_not_installed=True, yes=True,
                **kwargs) -> None:
        """Upgrades packages from various sources."""
        if not pkg_ids:
            raise Chocolatey.TypeError("upgrade() "
                                       "missing at least 1 required positional argument")
        self._omit_args(kwargs, "install_if_not_installed", "yes")  # , "verbose")
        try:
            self.cmd.upgrade(*pkg_ids,
                             install_if_not_installed=install_if_not_installed,
                             yes=yes, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def uninstall(self, *pkg_ids, yes=True, **kwargs) -> None:
        """Uninstalls packages."""
        if not pkg_ids:
            raise Chocolatey.TypeError("uninstall() "
                                       "missing at least 1 required positional argument")
        self._omit_args(kwargs, "yes")  # , "verbose")
        try:
            self.cmd.uninstall(*pkg_ids, yes=yes, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def pinned(self, **kwargs) -> Dict[str, List[Package]]:
        """Retrieves a list of packages suppress for upgrades."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.pin("list", limit_output=True,
                                  **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._packages(output.stdout)

    def pin_add(self, *, pkg_id: str, **kwargs) -> None:
        """Suppress upgrades for a package."""
        self._omit_args(kwargs, "reason")  # , "verbose") # --reason don't work
        try:
            self.cmd.pin("add", name=pkg_id,
                         # reason=reason if reason is False else f'"{reason}"',
                         **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def pin_remove(self, *, pkg_id: str, **kwargs) -> None:
        """Remove suppressing of upgrades for a package."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.pin("remove", name=pkg_id, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def pack(self, nuspec_file_path=False, *, output_directory=False,
             **kwargs) -> None:
        """Packages nuspec, scripts, and other package resources into a nupkg file."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            arg = [nuspec_file_path] if nuspec_file_path is not False else []
            self.cmd.pack(*arg, output_directory=output_directory, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def push(self, nupkg_file_path=False, yes=True, **kwargs) -> None:  # pragma: no cover
        """Pushes a compiled nupkg to a source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            arg = [nupkg_file_path] if nupkg_file_path is not False else []
            self.cmd.push(*arg, yes=yes, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def new_package(self, *, pkg_id: str,
                    properties: Dict[str, str] | None = None,
                    **kwargs) -> None:
        """Creates template files for creating a new Chocolatey package."""
        # <name> [<options/switches>] [<property=value> <propertyN=valueN>]
        self._omit_args(kwargs)  # , "verbose")
        props = [f'"{prop}={value}"' for prop, value in properties.items()]
        try:
            self.cmd.new(pkg_id, *props, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    # Configuration - https://docs.chocolatey.org/en-us/configuration

    def config(self, **kwargs) ->  Dict[str, Config]:
        """Retrieve config settings."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.config("list", limit_output=True,
                                     **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._config(output.stdout, klass=Config)

    def config_get(self, *, name: str, **kwargs) -> Union[str, bool]:
        """Get config value."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            output = self.cmd.config("get", name=name, limit_output=True,
                                     **self._capture_output, **kwargs)
            value = output.stdout
            if value and value[-1] == "\n": value = value[:-1]
            if value and value[-1] == "\r": value = value[:-1]
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return _str2bool("stdout", value, with_check=False)

    def config_set(self, *, name: str, value, **kwargs) -> None:
        """Set config value."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            if value is None or value == "":
                self.cmd.config("unset", name=name,
                                **self._capture_output, **kwargs)
            else:
                self.cmd.config("set", name=name, value=_bool2str(name, value),
                                **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def config_unset(self, *, name: str, **kwargs) -> None:
        """Unset config."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.config("unset", name=name,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def sources(self, **kwargs) -> Dict[str, Source]:
        """Retrieve default sources."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.source("list", limit_output=True,
                                     **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._config(output.stdout, klass=Source)

    def source_add(self, *, name: str, source: str, **kwargs) -> None:
        """Add source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.source("add", name=name, source=source,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def source_enable(self, *, name: str, **kwargs) -> None:
        """Enable source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.source("enable", name=name,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def source_disable(self, *, name: str, **kwargs) -> None:
        """Disable source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.source("disable", name=name,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def source_remove(self, *, name: str, **kwargs) -> None:
        """Remove source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.source("remove", name=name,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def features(self, **kwargs) -> Dict[str, Feature]:
        """Retrieve features."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.feature("list", limit_output=True,
                                      **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._config(output.stdout, klass=Feature)

    def feature_get(self, *, name: str, **kwargs) -> bool:
        """Get feature value."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            output = self.cmd.feature("get", name=name, limit_output=True,
                                      **self._capture_output, **kwargs)
            value = output.stdout.strip()
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return _str2bool("stdout", value, literals=("enabled", "disabled"))

    def feature_enable(self, *, name: str, **kwargs) -> None:
        """Enable feature."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.feature("enable", name=name,
                             **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def feature_disable(self, *, name: str, **kwargs) -> None:
        """Disable feature."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.feature("disable", name=name,
                             **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def apikeys(self, **kwargs) -> List[ApiKey]:
        """Retrieve the list of API keys."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.apikey("list", limit_output=True,
                                     **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return list(self._config(output.stdout, klass=ApiKey).values())

    def apikey_add(self, *, source: str, api_key: str, **kwargs) -> None:
        """Add API key for source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.apikey("add", source=source, api_key=api_key,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def apikey_remove(self, *, source: str, **kwargs) -> None:
        """Remove API key for source."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.apikey("remove", source=source,
                            **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def templates(self, **kwargs) -> Dict[str, Template]:
        """Retrieve templates."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.template("list", limit_output=True,
                                       **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return self._config(output.stdout, klass=Template)

    def template_info(self, *, name: str, **kwargs) -> Template:
        """Retrieve template info."""
        self._omit_args(kwargs, "limit_output", "verbose")
        try:
            output = self.cmd.template("info", name=name, limit_output=True,
                                       **self._capture_output, **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)
        return list(self._config(output.stdout, klass=Template).values())[0]

    def cache_list(self, **kwargs) -> None:
        """Displays information about the local HTTP caches used to store queries."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.cache("list", **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    def cache_remove(self, **kwargs) -> None:
        """Remove the local HTTP caches used to store queries."""
        self._omit_args(kwargs)  # , "verbose")
        try:
            self.cmd.cache("remove", **kwargs)
        except run.CalledProcessError as exc:
            self._handle_exception(exc)

    # ----- internals ----- #

    # OUTPUT_QUIET ERROR_QUIET
    # quiet_stdout=True,
    # quiet_stderr=True,
    # env=None,
    # out_cm=None,
    # err_cm=None,

    _capture_output = dict(text=True, capture_output=True)

    def _packages(self, out: str, *, klass=Package,
                  allow_multiple=None) -> Dict[str, Any]:
        lines = [line.strip() for line in out.strip().splitlines()]
        lines.sort(key=str.casefold)
        packages = defaultdict(list)
        for line in lines:
            # print("LINE:", line)
            package = klass(*line.split("|"))
            # print("PKG: ", package)
            if self._allow_multiple if allow_multiple is None else allow_multiple:
                packages[package.id].append(package)
            else:
                packages[package.id] = package
        return dict(packages)

    @classmethod
    def _info(cls, out: str,
              pkg_info: PackageInfo | None) -> PackageInfo | None:
        # if not found, returns None
        if not out.strip() or pkg_info is None:
            return None  # pragma: no cover

        out = out.replace("\r\n", "\n")

        # pre-parse
        out = re.sub(r"[\t ]*\|", "\n", out)
        out = re.sub(r"\n[\t ]*Package url[\t ]*\n", "\n Package url: n/a\n", out)
        # parse
        info_header_pattern = r"\s*Chocolatey[\t ]+v.+?([\t ]*\n)+" + \
                              rf"\s*{pkg_info.id}[\t ]+{pkg_info.version}([\t ]+\[\w+\])?" + \
                              r"([\t ]*\n)+"
        info_pattern        = r"(?P<info>(.*\n)*)"
        info_footer_pattern = r"\n(?P<pkgs_found>\d+)" + \
                              r"[\t ]+packages[\t ]+(found|installed)[\t ]*\.([\t ]*\n)*"
        match = re.match(info_header_pattern
                         + info_pattern
                         + info_footer_pattern, out)
        info_text = textwrap.dedent(match.captures("info")[0])
        if not info_text.endswith("\n"): info_text += "\n"

        info_key_pattern   = r"(?P<info_key>[^\t :]+([\t ]+[^\t :]+)*)"
        info_value_pattern = r"(?P<info_value>.*(\n([\t ]+.*)?)*)"
        info_item_pattern  = rf"{info_key_pattern}[\t ]*:[\t ]*{info_value_pattern}"
        info_items_pattern = rf"(?P<info_items>({info_item_pattern}\n)+)"

        match = re.match(info_items_pattern, info_text)

        info = NocaseDict(zip(match.captures("info_key"),
                              match.captures("info_value")))
        # print("============================")
        # print(dict(info))
        # print("============================")
        pkg_info.description = info.pop("Description", "")
        pkg_info.title       = info.pop("Title", "")
        pkg_info.summary     = info.pop("Summary", "")
        # additional
        pkg_info.published   = info.pop("Published", "")

        # TODO:
        """
        Exist in package with empty metadata (only required fields set).
        ----------------------------------------------------------------
        Number of Downloads: n/a
        Downloads for this version: n/a
        Package url
        Chocolatey Package Source: n/a
        Tags:
        Software Site: n/a
        Software License: n/a
        """
        return pkg_info

    @classmethod
    def _config(cls, out: str, *, klass) -> Dict[str, Any]:
        lines = [line.strip() for line in out.strip().splitlines()]
        lines.sort(key=str.casefold)
        configs = {}
        for line in lines:
            fields = line.split("|")
            configs[fields[0]] = klass(*fields)
        return configs

    @classmethod
    def _omit_args(cls, kwargs, *unnecessary):
        for omit in unnecessary:
            kwargs.pop(omit, None)

    def _handle_exception(self, exc, **kwargs):
        raise exc  # pragma: no cover
        # raise Chocolatey.RuntimeError(???)

    class Error(Exception):
        """Chocolatey error."""

    class TypeError(TypeError, Error):  # noqa: A001
        """Chocolatey type error."""

    class ValueError(ValueError, Error):  # noqa: A001
        """Chocolatey value error."""

    class RuntimeError(RuntimeError, Error):  # noqa: A001
        """Chocolatey runtime error."""


def _bool2str(name: str, value: Any, *,
              literals=("true", "false")):
    if not isinstance(value, bool):
        return value
    return literals[0 if value else 1]


def _str2bool(name: str, value: Any, *, with_check=True,
              literals=("true", "false")):
    if not isinstance(value, str):
        return value
    if not (value.lower() in literals):
        if not with_check: return value
        raise Chocolatey.ValueError(f"Improper value ({value}) for '{name}' attribute!")
    return (value.lower() == literals[0])


def _str2int(name: str, value: Any, *, with_check=True):
    if not isinstance(value, str):
        return value
    if not (value.isdigit()
            or value[1:].isdigit() and value[1:] in ("+", "-")):
        if not with_check: return value
        raise Chocolatey.ValueError(f"Improper value ({value}) for '{name}' attribute!")
    return int(value)


def _str2none(name: str, value: Any):
    return None if isinstance(value, str) and not value else value



# Copyright (c) 2022 Adam Karpierz
# SPDX-License-Identifier: Zlib

from .__about__ import * ; del __about__  # noqa

from ._chocolatey     import * ; del _chocolatey  # noqa
from ._chocolatey_cmd import * ; del _chocolatey_cmd  # noqa



from pathlib import Path
import logging
# logging.basicConfig(level=logging.INFO)
# logging.basicConfig(level=logging.DEBUG)

here = Path(__file__).resolve().parent

from chocolatey import Chocolatey
choco = Chocolatey()

print("choco version:", choco.version)

(here/"help.txt").open("w", newline="").write(choco.help(command="help"))
for command in ("apikey", "cache", "config", "export", "feature", "info",
                "install", "list", "new", "outdated", "pack", "pin", "push",
                "rule", "search", "source", "template", "uninstall", "upgrade"):
    fpath = here/f"help_{command}.txt"
    fpath.open("w", newline="").write(choco.help(command=command))



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

__all__ = ('run', 'split_kwargs')

from typing import Any, Sequence, Tuple, Dict
import subprocess
import logging

log = logging.getLogger(__name__)


def run(*args, start_terminal_window=False, **kwargs):
    """
    Run the command described by args.

    Wait for command to complete, then return a subprocess.CompletedProcess instance.
    :param args: command line arguments which are provided to E-Sys.bat
    :param start_terminal_window: start E-Sys.bat in separate console window (for server mode)
    :param kwargs: arguments tu control subprocess execution
    :return: subprocess.CompletedProcess instance
    """
    if start_terminal_window:  # pragma: no cover
        args = ["cmd.exe", "/C", "start", *args]
    output = subprocess.run([str(arg) for arg in args], check=kwargs.pop("check", True), **kwargs)
    print_cmd = [("*****" if isinstance(arg, run.SafeString) else arg) for arg in args]
    log.debug(f"cmd:{print_cmd}, returncode:{output.returncode}")
    return output


run.CompletedProcess = subprocess.CompletedProcess

run.PIPE    = subprocess.PIPE
run.STDOUT  = subprocess.STDOUT
run.DEVNULL = subprocess.DEVNULL

run.SubprocessError    = subprocess.SubprocessError
run.TimeoutExpired     = subprocess.TimeoutExpired
run.CalledProcessError = subprocess.CalledProcessError

run.SafeString = type("SafeString", (str,), dict())


def split_kwargs(kwargs: Dict[str, Any], forbidden_kwargs: Sequence[str]) \
        -> Tuple[Dict[str, Any], Dict[str, Any]]:
    allowed_kwargs  = {key: val for key, val in kwargs.items()
                       if key not in forbidden_kwargs}
    reserved_kwargs = {key: val for key, val in kwargs.items()
                       if key in forbidden_kwargs}
    return (allowed_kwargs, reserved_kwargs)



# Copyright (c) 2022 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import Any, Optional, Union, Sequence, Tuple, List, Dict
import unittest
from functools import partial
from pathlib import Path
import os, shutil, tempfile
import threading

from rich.pretty import pprint
pprint = partial(pprint, max_length=500)
from ._util import pushd

import chocolatey
from chocolatey import Chocolatey

here = Path(__file__).resolve().parent
data_dir = here/"data"


def setUpModule():
    Chocolatey.setup()


class ChocolateyTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.choco = Chocolatey()
        cls.lock = threading.Lock()

    def setUp(self):
        self.lock.acquire()

    def tearDown(self):
        self.lock.release()

    ### High-level API ###

    def test_version(self):
        """Gets the Chocolatey version."""
        version = self.choco.version
        self.assertIsInstance(version, str)
        print(flush=True)
        print("CHOCO VERSION:", version)

    def test_version_info(self):
        """Gets the Chocolatey version info."""
        version_info = self.choco.version_info
        self.assertIsInstance(version_info, chocolatey.version_info)
        self.assertIsInstance(version_info.major,  int)
        self.assertIsInstance(version_info.minor,  int)
        self.assertIsInstance(version_info.micro,  int)
        self.assertIsInstance(version_info.serial, int)

    def test_help(self):
        """Gets the help information for choco and choco commands."""
        help_str = self.choco.help()
        self.assertIsInstance(help_str, str)
        help_str = self.choco.help(command="search")
        self.assertIsInstance(help_str, str)

    def test_installed(self):
        """Retrieves a list of locally installed packages."""
        installed = self.choco.installed()
        self.assertIsInstance(installed, Dict)
        for pkg_id, package in installed.items():
            self.assertIsInstance(pkg_id, str)
            self.assertIsInstance(package, chocolatey.Package)
        pkg_id = "chocolatey"
        self.assertIsInstance(installed[pkg_id], chocolatey.Package)

    def test_outdated(self):
        """Retrieves information about packages that are outdated."""
        outdated = self.choco.outdated()
        self.assertIsInstance(outdated, Dict)
        for pkg_id, package in outdated.items():
            self.assertIsInstance(pkg_id, str)
            self.assertIsInstance(package, chocolatey.PackageOutdated)

    def test_search(self):
        """Searches remote packages."""
        pkg_id = "chocolatey"
        # exact
        found = self.choco.search(pkg_id, exact=True)
        self.assertIsInstance(found, Dict)
        self.assertIsInstance(found[pkg_id], chocolatey.Package)
        found = self.choco.search(pkg_id, exact=True, all_versions=True)
        self.assertIsInstance(found, Dict)
        self.assertIsInstance(found[pkg_id], List)
        self.assertIsInstance(found[pkg_id][0], chocolatey.Package)
        # as filter
        found = self.choco.search(pkg_id, by_id_only=True)
        self.assertIsInstance(found, Dict)
        self.assertIsInstance(list(found.values())[0], chocolatey.Package)
        found = self.choco.search(pkg_id, by_id_only=True, all_versions=True)
        self.assertIsInstance(found, Dict)
        self.assertIsInstance(list(found.values())[0], List)
        self.assertIsInstance(list(found.values())[0][0], chocolatey.Package)

    def test_info(self):
        """Retrieves package information."""
        pkg_id = "chocolatey"
        pkg_info = self.choco.info(pkg_id=pkg_id)
        self.assertIsInstance(pkg_info, chocolatey.PackageInfo)

    def test_export(self):
        """Exports list of currently installed packages."""
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)
            pkg_config    = temp_dir/"packages.config"
            pkg_config0   = temp_dir/"packages0.config"
            pkg_config0v  = temp_dir/"packages0v.config"
            pkg_config0nv = temp_dir/"packages0nv.config"
            pkg_config1   = temp_dir/"packages1.config"
            pkg_config1v  = temp_dir/"packages1v.config"
            pkg_config1nv = temp_dir/"packages1nv.config"
            pkg_config2   = temp_dir/"packages2.config"
            pkg_config2v  = temp_dir/"packages2v.config"
            pkg_config2nv = temp_dir/"packages2nv.config"
            with pushd(temp_dir):
                self.choco.export()
                shutil.move(pkg_config, pkg_config0)
                self.choco.export(include_version=True)
                shutil.move(pkg_config, pkg_config0v)
                self.choco.export(include_version=False)
                shutil.move(pkg_config, pkg_config0nv)
            self.choco.export(pkg_config1)
            self.choco.export(pkg_config1v,  include_version=True)
            self.choco.export(pkg_config1nv, include_version=False)
            self.choco.export(output_file_path=pkg_config2)
            self.choco.export(output_file_path=pkg_config2v,  include_version=True)
            self.choco.export(output_file_path=pkg_config2nv, include_version=False)
            self.assertTrue(pkg_config0.exists())
            self.assertTrue(pkg_config0v.exists())
            self.assertTrue(pkg_config0nv.exists())
            self.assertTrue(pkg_config1.exists())
            self.assertTrue(pkg_config1v.exists())
            self.assertTrue(pkg_config1nv.exists())
            self.assertTrue(pkg_config2.exists())
            self.assertTrue(pkg_config2v.exists())
            self.assertTrue(pkg_config2nv.exists())

    def test_install(self):
        """Installs packages using configured sources."""
        print(flush=True)
        pkg_id = "py-chocolatey.Test1"
        if pkg_id in self.choco.installed(): self.choco.uninstall(pkg_id)
        installed_before = self.choco.installed()
        try:
            self.choco.install(pkg_id, source=data_dir)
            installed_after = self.choco.installed()
            diff_keys = set(installed_after) - set(installed_before)
            self.assertEqual(diff_keys, set([pkg_id]))
        finally:
            if pkg_id in self.choco.installed(): self.choco.uninstall(pkg_id)
        with self.assertRaises(TypeError):
            self.choco.install(source=data_dir)

    def test_upgrade(self):
        """Upgrades packages from various sources."""
        print(flush=True)
        pkg_id = "py-chocolatey.Test2"
        if pkg_id in self.choco.installed(): self.choco.uninstall(pkg_id)
        installed_before = self.choco.installed()
        try:
            self.choco.upgrade(pkg_id, source=data_dir)
            installed_after = self.choco.installed()
            diff_keys = set(installed_after) - set(installed_before)
            self.assertEqual(diff_keys, set([pkg_id]))
        finally:
            if pkg_id in self.choco.installed(): self.choco.uninstall(pkg_id)
        with self.assertRaises(TypeError):
            self.choco.upgrade(source=data_dir)

    def test_uninstall(self):
        """Uninstalls packages."""
        print(flush=True)
        pkg_id = "py-chocolatey.Test3"
        try:
            self.choco.install(pkg_id, source=data_dir)
            installed_before = self.choco.installed()
            self.choco.uninstall(pkg_id)
            installed_after = self.choco.installed()
            diff_keys = set(installed_before) - set(installed_after)
            self.assertEqual(diff_keys, set([pkg_id]))
        except:  # pragma: no cover
            if pkg_id in self.choco.installed(): self.choco.uninstall(pkg_id)
        with self.assertRaises(TypeError):
            self.choco.uninstall()

    def test_pinned(self):
        """Retrieves a list of packages suppress for upgrades."""
        pinned = self.choco.pinned()
        self.assertIsInstance(pinned, Dict)
        for pkg_id, package in pinned.items():
            self.assertIsInstance(pkg_id, str)
            self.assertIsInstance(package, chocolatey.Package)

    def test_pin_add(self):
        """Suppress upgrades for a package."""
        print(flush=True)
        pkg_id = "chocolatey"
        pinned_saved = self.choco.pinned()
        self.choco.pin_remove(pkg_id=pkg_id)
        pinned_before = self.choco.pinned()
        self.choco.pin_add(pkg_id=pkg_id)
        pinned_after = self.choco.pinned()
        self.assertIsInstance(pinned_before, Dict)
        self.assertIsInstance(pinned_after,  Dict)
        diff_keys = set(pinned_after) - set(pinned_before)
        self.assertEqual(diff_keys, set([pkg_id]))
        self.choco.pin_remove(pkg_id=pkg_id)

    def test_pin_remove(self):
        """Remove suppressing of upgrades for a package."""
        print(flush=True)
        pkg_id = "chocolatey"
        pinned_saved = self.choco.pinned()
        self.choco.pin_add(pkg_id=pkg_id)
        pinned_before = self.choco.pinned()
        self.choco.pin_remove(pkg_id=pkg_id)
        pinned_after = self.choco.pinned()
        self.assertIsInstance(pinned_before, Dict)
        self.assertIsInstance(pinned_after,  Dict)
        diff_keys = set(pinned_before) - set(pinned_after)
        self.assertEqual(diff_keys, set([pkg_id]))

    def test_pack(self):
        """Packages nuspec, scripts, and other Chocolatey package resources
        into a nupkg file."""
        print(flush=True)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)
            with pushd(temp_dir):
                pkg4_id = "py-chocolatey.Test4"
                pkg5_id = "py-chocolatey.Test5"
                self.choco.pack(data_dir/pkg4_id/(pkg4_id + ".nuspec"))
                self.choco.pack(nuspec_file_path=data_dir/pkg5_id/(pkg5_id + ".nuspec"))
                self.assertTrue(Path(pkg4_id + ".1.0.4.nupkg").exists())
                self.assertTrue(Path(pkg5_id + ".1.0.5.nupkg").exists())

    def test_push(self):
        """Pushes a compiled nupkg to a source."""
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)
            with pushd(temp_dir):
                pkg_id = "py-chocolatey.Test0"
                self.choco.pack(nuspec_file_path=data_dir/pkg_id/(pkg_id + ".nuspec"))
                self.assertTrue(Path(pkg_id + ".1.0.0.nupkg").exists())

    def test_new_package(self):
        """Creates template files for creating a new Chocolatey package."""
        print(flush=True)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)
            pkg_id = "py-chocolatey.Test0"
            self.choco.new_package(pkg_id=pkg_id,  auto=True, version="1.2.0",
                                   outputdirectory=temp_dir,
                                   properties = dict(maintainername="This guy"))
            self.assertTrue((temp_dir/pkg_id).is_dir())
            self.assertTrue((temp_dir/pkg_id/(pkg_id + ".nuspec")).is_file())
            self.assertTrue((temp_dir/pkg_id/"_TODO.txt").is_file())
            self.assertTrue((temp_dir/pkg_id/"ReadMe.md").is_file())
            self.assertTrue((temp_dir/pkg_id/"tools").is_dir())
            self.assertTrue((temp_dir/pkg_id/"tools/chocolateybeforemodify.ps1").is_file())
            self.assertTrue((temp_dir/pkg_id/"tools/chocolateyinstall.ps1").is_file())
            self.assertTrue((temp_dir/pkg_id/"tools/chocolateyuninstall.ps1").is_file())
            self.assertTrue((temp_dir/pkg_id/"tools/LICENSE.txt").is_file())
            self.assertTrue((temp_dir/pkg_id/"tools/VERIFICATION.txt").is_file())

    def test_config(self):
        """Retrieve config settings."""
        configs = self.choco.config()
        self.assertIsInstance(configs, Dict)
        for cfg_id, config in configs.items():
            self.assertIsInstance(cfg_id, str)
            self.assertIsInstance(config, chocolatey.Config)

    def test_config_get(self):
        """Get config value."""
        self.choco.config_unset(name="TEST1")
        self.choco.config_set(name="TEST1", value="TEST1_VALUE1")
        value = self.choco.config_get(name="TEST1")
        self.assertEqual(value, "TEST1_VALUE1")
        self.choco.config_set(name="TEST1", value="TEST1_VALUE2")
        value = self.choco.config_get(name="TEST1")
        self.assertEqual(value, "TEST1_VALUE2")
        self.choco.config_unset(name="TEST1")

    def test_config_set(self):
        """Set config value."""
        self.choco.config_unset(name="TEST2")
        self.choco.config_set(name="TEST2", value="TEST2_VALUE1")
        value = self.choco.config_get(name="TEST2")
        self.assertEqual(value, "TEST2_VALUE1")
        self.choco.config_set(name="TEST2", value="TEST2_VALUE2")
        value = self.choco.config_get(name="TEST2")
        self.assertEqual(value, "TEST2_VALUE2")
        self.choco.config_set(name="TEST2", value=True)
        value = self.choco.config_get(name="TEST2")
        self.assertTrue(value)
        self.choco.config_set(name="TEST2", value=None)
        value = self.choco.config_get(name="TEST2")
        self.assertIs(value, "")
        self.choco.config_set(name="TEST2", value="")
        value = self.choco.config_get(name="TEST2")
        self.assertIs(value, "")
        self.choco.config_unset(name="TEST2")

    def test_config_unset(self):
        """Unset config."""
        self.choco.config_unset(name="TEST3")
        self.choco.config_set(name="TEST3", value="TEST3_VALUE1")
        value = self.choco.config_get(name="TEST3")
        self.assertEqual(value, "TEST3_VALUE1")
        self.choco.config_unset(name="TEST3")
        #with self.assertRaises(Exception):
        value = self.choco.config_get(name="TEST3")

    def test_sources(self):
        """Retrieve default sources."""
        sources = self.choco.sources()
        self.assertIsInstance(sources, Dict)
        for name, source in sources.items():
            self.assertIsInstance(source, chocolatey.Source)

    def test_source_add(self):
        """Add source."""
        source_name = "_test_source"
        sources_before = self.choco.sources()
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            sources_after = self.choco.sources()
            diff_keys = set(sources_after) - set(sources_before)
            self.assertEqual(diff_keys, set([source_name]))
        finally:
            self.choco.source_remove(name=source_name)

    def test_source_enable(self):
        """Enable source."""
        source_name = "_test_source"
        sources_before = self.choco.sources()
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            sources_after = self.choco.sources()
            diff_keys = set(sources_after) - set(sources_before)
            self.assertEqual(diff_keys, set([source_name]))
            self.choco.source_enable(name=source_name)
        finally:
            self.choco.source_remove(name=source_name)

    def test_source_disable(self):
        """Disable source."""
        source_name = "_test_source"
        sources_before = self.choco.sources()
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            sources_after = self.choco.sources()
            diff_keys = set(sources_after) - set(sources_before)
            self.assertEqual(diff_keys, set([source_name]))
            self.choco.source_disable(name=source_name)
        finally:
            self.choco.source_remove(name=source_name)

    def test_source_remove(self):
        """Remove source."""
        source_name = "_test_source"
        sources_before = self.choco.sources()
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            sources_after = self.choco.sources()
            diff_keys = set(sources_after) - set(sources_before)
            self.assertEqual(diff_keys, set([source_name]))
        finally:
            self.choco.source_remove(name=source_name)
        sources_after = self.choco.sources()
        self.assertEqual(set(sources_after), set(sources_before))

    def test_features(self):
        """Retrieve features."""
        features = self.choco.features()
        self.assertIsInstance(features, Dict)
        for name, feature in features.items():
            self.assertIsInstance(feature, chocolatey.Feature)

    def test_feature_get(self):
        """Get feature value."""
        feature_name = "autoUninstaller"
        feature = self.choco.feature_get(name=feature_name)
        self.assertIsInstance(feature, bool)

    def test_feature_enable(self):
        """Enable feature."""
        feature_name = "autoUninstaller"
        feature_org = self.choco.feature_get(name=feature_name)
        self.assertIsInstance(feature_org, bool)
        try:
            self.choco.feature_disable(name=feature_name)
            feature = self.choco.feature_get(name=feature_name)
            self.assertIsInstance(feature, bool)
            self.assertFalse(feature)
            # enable feature
            self.choco.feature_enable(name=feature_name)
            feature = self.choco.feature_get(name=feature_name)
            self.assertIsInstance(feature, bool)
            self.assertTrue(feature)
        finally:
            (self.choco.feature_enable
             if feature_org else
             self.choco.feature_disable)(name=feature_name)

    def test_feature_disable(self):
        """Disable feature."""
        feature_name = "autoUninstaller"
        feature_org = self.choco.feature_get(name=feature_name)
        self.assertIsInstance(feature_org, bool)
        try:
            self.choco.feature_enable(name=feature_name)
            feature = self.choco.feature_get(name=feature_name)
            self.assertIsInstance(feature, bool)
            self.assertTrue(feature)
            # disable feature
            self.choco.feature_disable(name=feature_name)
            feature = self.choco.feature_get(name=feature_name)
            self.assertIsInstance(feature, bool)
            self.assertFalse(feature)
        finally:
            (self.choco.feature_enable
             if feature_org else
             self.choco.feature_disable)(name=feature_name)

    def test_apikeys(self):
        """Retrieve the list of API keys."""
        source_name = "_test_source"
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            api_key = "123-123123-123"
            self.choco.apikey_add(source=data_dir, api_key=api_key)
            try:
                apikeys = self.choco.apikeys()
                self.assertIsInstance(apikeys, List)
                for apikey in apikeys:
                    self.assertIsInstance(apikey, chocolatey.ApiKey)
            finally:
                self.choco.apikey_remove(source=data_dir)
        finally:
            self.choco.source_remove(name=source_name)

    def test_apikey_add(self):
        """Add API key for source."""
        source_name = "_test_source"
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            api_key = "123-123123-123"
            apikeys_before = self.choco.apikeys()
            self.choco.apikey_add(source=data_dir, api_key=api_key)
            try:
                apikeys_after = self.choco.apikeys()
                diff_lists = [item for item in apikeys_after
                              if item not in apikeys_before]
                self.assertTrue(bool(diff_lists))
                self.assertEqual(diff_lists[0].source, str(data_dir))
                self.assertEqual(diff_lists[0].info, "(Authenticated)")
            finally:
                self.choco.apikey_remove(source=data_dir)
        finally:
            self.choco.source_remove(name=source_name)

    def test_apikey_remove(self):
        """Remove API key for source."""
        apikeys_before = self.choco.apikeys()
        #self.choco.apikey_remove(source="")
        apikeys_after = self.choco.apikeys()

        source_name = "_test_source"
        self.choco.source_add(name=source_name, source=data_dir)
        try:
            api_key = "123-123123-123"
            self.choco.apikey_add(source=data_dir, api_key=api_key)
            apikeys_before = self.choco.apikeys()
            self.choco.apikey_remove(source=data_dir)
            apikeys_after = self.choco.apikeys()
            diff_lists = [item for item in apikeys_before
                          if item not in apikeys_after]
            self.assertTrue(bool(diff_lists))
            self.assertEqual(diff_lists[0].source, str(data_dir))
            self.assertEqual(diff_lists[0].info, "(Authenticated)")
        finally:
            self.choco.source_remove(name=source_name)

    def test_templates(self):
        """Retrieve templates."""
        templates = self.choco.templates()
        self.assertIsInstance(templates, Dict)
        for name, template in templates.items():
            self.assertIsInstance(name, str)
            self.assertIsInstance(template, chocolatey.Template)

    def test_template_info(self):
        """Retrieve template info."""
        #template_name = "py-chocolatey.test_template"
        template_name = "built-in"
        template = self.choco.template_info(name=template_name)
        self.assertIsInstance(template, chocolatey.Template)

    def test_cache_list(self):
        """Displays information about the local HTTP caches used to store
        queries."""
        print(flush=True)
        self.choco.cache_list()

    def test_cache_remove(self):
        """Remove the local HTTP caches used to store queries."""
        print(flush=True)
        self.choco.cache_remove(expired=True)
        self.choco.cache_remove()

    def test_miscs(self):
        """Test internal utilities."""
        # _bool2str
        from chocolatey._chocolatey import _bool2str
        self.assertEqual(_bool2str("xyz", True), "true")
        self.assertEqual(_bool2str("xyz", False), "false")
        self.assertEqual(_bool2str("xyz", "abc"), "abc")
        self.assertEqual(_bool2str("xyz", 1), 1)
        self.assertEqual(_bool2str("xyz", 123), 123)
        # _str2bool
        from chocolatey._chocolatey import _str2bool
        self.assertEqual(_str2bool("xyz", "true", with_check=False), True)
        self.assertEqual(_str2bool("xyz", "false", with_check=False), False)
        self.assertEqual(_str2bool("xyz", "abc", with_check=False), "abc")
        self.assertEqual(_str2bool("xyz", True, with_check=False), True)
        self.assertEqual(_str2bool("xyz", False, with_check=False), False)
        self.assertEqual(_str2bool("xyz", 1, with_check=False), 1)
        self.assertEqual(_str2bool("xyz", 123, with_check=False), 123)
        self.assertEqual(_str2bool("xyz", "true"), True)
        self.assertEqual(_str2bool("xyz", "false"), False)
        with self.assertRaises(Chocolatey.ValueError):
            _str2bool("xyz", "abc")
        self.assertEqual(_str2bool("xyz", True), True)
        self.assertEqual(_str2bool("xyz", False), False)
        self.assertEqual(_str2bool("xyz", 1), 1)
        self.assertEqual(_str2bool("xyz", 123), 123)
        # _str2int
        from chocolatey._chocolatey import _str2int
        self.assertEqual(_str2int("xyz", "123", with_check=False), 123)
        self.assertEqual(_str2int("xyz", "abc", with_check=False), "abc")
        self.assertEqual(_str2int("xyz", True, with_check=False), True)
        self.assertEqual(_str2int("xyz", False, with_check=False), False)
        self.assertEqual(_str2int("xyz", 1, with_check=False), 1)
        self.assertEqual(_str2int("xyz", 123, with_check=False), 123)
        self.assertEqual(_str2int("xyz", "123"), 123)
        with self.assertRaises(Chocolatey.ValueError):
            _str2int("xyz", "abc")
        self.assertEqual(_str2int("xyz", True), True)
        self.assertEqual(_str2int("xyz", False), False)
        self.assertEqual(_str2int("xyz", 1), 1)
        self.assertEqual(_str2int("xyz", 123), 123)
        # _str2none
        from chocolatey._chocolatey import _str2none
        self.assertEqual(_str2none("xyz", ""), None)
        self.assertEqual(_str2none("xyz", None), None)
        self.assertEqual(_str2none("xyz", "abc"), "abc")
        self.assertEqual(_str2none("xyz", True), True)
        self.assertEqual(_str2none("xyz", False), False)
        self.assertEqual(_str2none("xyz", 1), 1)
        self.assertEqual(_str2none("xyz", 123), 123)
